import astra
import numpy as np
from os import mkdir
from os.path import isdir
from PIL import Image

def project_from_2D(phantom_id, vol_geom, n_projections, 
                    n_detectors, detector_spacing, angles, 
                    noise_factor=None, save_dir=None, use_gpu=False):
        """ Creates projection for the given input data.
            
            Parameters:
                - phantom_data: numpy array containing the 2D phantom.
                - n_projections: defines the number of projections 
                    (number of angles from where to sample measurements 
                - detector_spacing: pixel size
                - angles:
                - noise_factor:
                - save_dir: path of the directory to save image representation 
                    of projections, when defined. To be passed as a string.
                - use_gpu: (boolean) set to True to use gpu.

            Returns:
                projector_id, sinogram_id and sinogram matrix
        """

        # create projection geometry
        proj_geom = astra.create_proj_geom('parallel', detector_spacing, 
                                            n_detectors, angles)
        # choose projector
        if use_gpu:
            proj_id = astra.create_projector('cuda', proj_geom, vol_geom)
        else:
            proj_id = astra.create_projector('linear', proj_geom, vol_geom)
        sino_id, sinogram = astra.creators.create_sino(phantom_id, proj_id)
        # Apply Poisson noise.
        if noise_factor != None:
            sinogram += np.random.poisson(lam=noise_factor, size=sinogram.shape)
            sino_id = astra.data2d.create('-sino', proj_geom, sinogram)
        # Save projections as images, if directory has been defined.
        if save_dir != None:
            if save_dir[-1] != '/':
                save_dir += '/'
            if not isdir(save_dir):
                mkdir(save_dir)
            proj_for_img = np.round(sinogram * (2**8- 1)).astype(np.uint8)
            for i in range(n_projections):
                Image.fromarray(proj_for_img[i]).save(save_dir+f'proj_{i}.png')

        return proj_id, sino_id, sinogram


import astra

def SIRT(vol_geom, vol_data, sino_id, iters=2000, use_gpu=False):
        # create starting reconstruction
        rec_id = astra.data2d.create('-vol', vol_geom, data=vol_data)
        # define SIRT config params
        alg_cfg = astra.astra_dict('SIRT_CUDA' if use_gpu else 'SIRT')
        alg_cfg['ProjectionDataId'] = sino_id
        alg_cfg['ReconstructionDataId'] = rec_id
        alg_cfg['option'] = {}
        alg_cfg['option']['MinConstraint'] = 0
        alg_cfg['option']['MaxConstraint'] = 255
        # define algorithm
        alg_id = astra.algorithm.create(alg_cfg)
        # run the algorithm
        astra.algorithm.run(alg_id, iters)
        # create reconstruction data
        rec = astra.data2d.get(rec_id)

        return rec_id, rec


import astra

def FBP(vol_geom, vol_data, projector_id, sino_id, iters=2000, use_gpu=False):
    # create starting reconstruction
    rec_id = astra.data2d.create('-vol', vol_geom, data=vol_data)
    # define SART configuration parameters
    alg_cfg = astra.astra_dict('FBP_CUDA' if use_gpu else 'FBP')
    alg_cfg['ProjectorId'] = projector_id
    alg_cfg['ProjectionDataId'] = sino_id
    alg_cfg['ReconstructionDataId'] = rec_id
    alg_cfg['option'] = {}
    alg_cfg['option']['MinConstraint'] = 0
    alg_cfg['option']['MaxConstraint'] = 255
    # define algorithm
    algorithm_id = astra.algorithm.create(alg_cfg)
    # run the algirithm
    astra.algorithm.run(algorithm_id, iters)
    # create reconstruction data
    rec = astra.data2d.get(rec_id)

    return rec_id, rec


from copy import deepcopy
import numpy as np
import astra
from scipy.ndimage import gaussian_filter

class DART():
    def __init__(self, gray_levels, p, rec_shape, 
                proj_geom, projector_id, sinogram):
        """ Instanciate DART with thw following parameters
            Parameters:
                - gray_levels: gray levels known a priori used in the segmentation step.
                - p: probability of a pixel to not be sampled as a free pixel.
                - rec_shape: shape of the volume to create as output.
                - proj_geom: projection geometry to use for the sinogram creation. 
                - projector_id: reference to the astra toolbox projector used to make the projections.
                - sinogram: sinogram as numpy matrix
        """
        self.gray_levels = gray_levels
        # define thresholds for gray levels with start and end values
        self.thresholds = self.update_gray_thresholds()
        self.p = p
        self.c, self.probs = [0,1], [self.p, 1-self.p]
        self.rec_shape = rec_shape
        self.vol_geom = astra.creators.create_vol_geom(self.rec_shape)
        # calculated in advance for efficiency
        self.all_neighbours_idx = [[self.pixel_neighborhood(rec_shape, i,j)
                                    for j in range(self.rec_shape[1])]
                                        for i in range(self.rec_shape[0])]
        self.proj_geom = proj_geom
        self.projector_id = projector_id
        self.sinogram = sinogram
        self.sinogram_id = astra.data2d.create('-sino', proj_geom, sinogram)

    def run(self, iters, p=None, gray_levels=None, 
            rec_alg="SART_CUDA", rec_iter=5):
        """ Parameters:
                - iters: (int) number of DART iteration to perform
                - p: (float) probability of a pixel to not be sampled as a free pixel.
                - gray_levels: (list) gray levels known a priori used in the segmentation step.
                - rec_algs: (string) tuple containing the initial and the iterated 
                    reconstruction algorithms to use.
                - rec_iters: (int) number of iterations of the reconstruction subrutine.
            Output:
                (np.array) returns the reconstructed phantom 
                of shape = vol_shape, as a numpy 2D array
        """
        # to run experiments on different gray values
        # and fixed pixel probabilities
        # without reinstanciating DART
        if p is not None:
            self.p = p
        if gray_levels is not None:
            self.gray_levels = gray_levels
            self.thresholds = self.update_gray_thresholds()
        # reconstruction algorithm check
        if rec_alg not in [ "SART", "SART_CUDA",
                            "SIRT", "SIRT_CUDA",
                            "FBP" , "FBP_CUDA"]:
            exit("Select a valid reconstruction algorithm.") 
        # create initial reconstruction
        curr_rec = self.ART(np.full(shape=self.rec_shape,fill_value=0.),
                                    mask=None, alg=rec_alg, iters=rec_iter)
        for i in range(iters):
            # segment current reconstructed image
            segmented_img = self.segment(curr_rec)
            # calculate boundary pixels
            boundary_pixels = self.boundary_pixels(segmented_img)
            # calculate free pixels
            free_pixels = self.free_pixels()
            # mask of all free pixels
            free_pixels = np.logical_or(boundary_pixels,free_pixels)
            # take indexes of non fixed pixels
            free_pixels_idx = np.where(free_pixels)
            # fixed pixels
            fixed_pixels = free_pixels == 0
            # take indexes of non fixed pixels
            fixed_pixels_idx = np.where(fixed_pixels)
            #create image to feed to reconstructor
            curr_rec[fixed_pixels_idx[0],
                     fixed_pixels_idx[1]] = segmented_img[fixed_pixels_idx[0],
                                                        fixed_pixels_idx[1]]
            # run reconstruction algorithm on free pixels
            curr_rec = self.ART(curr_rec, mask=free_pixels,
                                        alg=rec_alg, iters=rec_iter)
            # smoothing operation except on last iteration
            if i < iters - 1:
                smooth_rec = gaussian_filter(curr_rec, sigma=1)
                curr_rec[free_pixels_idx[0],
                            free_pixels_idx[1]] = smooth_rec[free_pixels_idx[0], 
                                                            free_pixels_idx[1]]
        return curr_rec

    def ART(self, rec, mask=None,
                    alg="SART_CUDA", iters= 5):
        """ Reconstruction with ARM techniques.
            Parameters:
                - rec: initial reconstructed image. (np.array)
                - mask: mask defining the pixels to update, (np.array)
                - alg: name of the reconstruction algorithm to use. (string)
                - iters: number of reconstruction iterations to perform. (int)
            Output:
                - reconstructed image. (np.array)
        """
        if mask is not None:
            # mask free pixels for the fixed pixel sinogram
            free_pixels_idx = np.where(mask > 0)
            fixed_rec = deepcopy(rec)
            fixed_rec[free_pixels_idx[0],
                    free_pixels_idx[1]] = 0
            # create fixed pixels' sinogram
            _, fixed_sino = astra.creators.create_sino(fixed_rec, self.projector_id)
            # create free pixels' sinogram
            free_sino = self.sinogram - fixed_sino
            free_sino_id = astra.data2d.create('-sino', self.proj_geom, free_sino)
            rec_id = astra.data2d.create('-vol', self.vol_geom, data=rec)
        else:  # first reconstrunction
            rec_id = astra.data2d.create('-vol', self.vol_geom, data=0.)
        # define configuration parameters
        alg_cfg = astra.astra_dict(alg)
        if alg_cfg != "SIRT" and alg_cfg != "SIRT_CUDA":
            alg_cfg['ProjectorId'] = self.projector_id
        alg_cfg['ProjectionDataId'] = free_sino_id if mask is not None else self.sinogram_id
        alg_cfg['ReconstructionDataId'] = rec_id
        alg_cfg['option'] = {}
        alg_cfg['option']['MinConstraint'] = 0
        alg_cfg['option']['MaxConstraint'] = 255
        if mask is not None:
            mask_id = astra.data2d.create('-vol', self.vol_geom, mask)
            alg_cfg['option']['ReconstructionMaskId'] = mask_id
        #define algorithm
        algorithm_id = astra.algorithm.create(alg_cfg)
        # run the algorithm
        astra.algorithm.run(algorithm_id, iters)
        # free memory
        astra.algorithm.delete(algorithm_id)
        # return the reconstructed values
        return astra.data2d.get(rec_id)

    def update_gray_thresholds(self):
        """ Updates algorithms' thresholds for the currently
            defined gray values.
        """
        return [0] + [(self.gray_levels[i]+self.gray_levels[i+1])/2 
                        for i in range(len(self.gray_levels)-1) ] + [255]

    def segment(self, img):
        """ Segments the input image to obtain an image with
            only the gray values specified. 
        """
        # compute segmentation
        segmented_img = np.full(img.shape, 0, dtype=np.uint8)
        for thresh_idx in range(len(self.thresholds)-1):
            cond = (img >= self.thresholds[thresh_idx]) * (img <= self.thresholds[thresh_idx+1])
            segmented_img[cond] = self.gray_levels[thresh_idx]
        return segmented_img

    def pixel_neighborhood(self, img_shape, x, y):
        """ Returns an array containing all the neighbours of the given pixel
        """
        # calculate all possible neighbours
        # related to the x,y coordinates
        neighbours = [(i,j) 
                        for i in range(x-1, x+2) 
                            if i > -1 and i < img_shape[0]
                                for j in range(y-1, y+2) 
                                    if j > -1 and j < img_shape[1] ] # and (x != i & y != j)
        return neighbours

    def boundary_pixels(self, img):
        """ Computes the boundary pixels of the image.
            Returns an image mask where boundary pixels 
            have value 1 and the rest all 0s.

            Parameters:
                - img: define the input image as a numpy array
        """
        # initialize output mask to 0
        bool_mask = np.full(fill_value=False, shape=img.shape[:2], dtype=np.bool8)
        for x in range(img.shape[0]):
            for y in range(img.shape[1]):
                pixel = img[x,y]
                # get actual neighbours values
                neighbours = [img[i] for i in self.all_neighbours_idx[x][y]]
                # update pixel value
                if np.any(neighbours != pixel):
                    bool_mask[x,y] = True
        return bool_mask

    def free_pixels(self):
        """ Computes the free pixels of the image.
            
            Parameters:
                - rec_shape: shape of the image for which 
                    to calculate the free pixels 

                - p: probability that a pixel is not sampled 
                    as a non boundary free pixel
        """
        free_pixels = np.random.choice(a=self.c, 
                                    size=self.rec_shape, 
                                    p=self.probs).astype(np.bool8)
        return free_pixels




import astra

def SART(vol_geom, vol_data, projector_id, sino_id, iters=2000, use_gpu=False):
        """ Simultaneous Algebraic Reconstruction Technique (SART) with
            randomized scheme. Used from DART as the continious update step.
        """
        # create starting reconstruction
        rec_id = astra.data2d.create('-vol', vol_geom, data=vol_data)
        # define SART configuration parameters
        alg_cfg = astra.astra_dict('SART_CUDA' if use_gpu else 'SART')
        alg_cfg['ProjectorId'] = projector_id
        alg_cfg['ProjectionDataId'] = sino_id
        alg_cfg['ReconstructionDataId'] = rec_id
        alg_cfg['option'] = {}
        alg_cfg['option']['MinConstraint'] = 0
        alg_cfg['option']['MaxConstraint'] = 255
        # define algorithm
        algorithm_id = astra.algorithm.create(alg_cfg)
        # run the algirithm
        astra.algorithm.run(algorithm_id, iters)
        # create reconstruction data
        rec = astra.data2d.get(rec_id)

        return rec_id, rec


import numpy as np
from PIL import Image
from os import makedirs
from os.path import exists

def create_phantoms(phantoms="semilunars",img_size=512, gray_values=[80,120,180], n=1, 
                        overlap=False, seed=None, img_name=None):
    generated = []
    if phantoms == "semilunars":
        generated = create_semilunars(img_size=img_size, gray_values=gray_values,
                                    n=n, overlap=overlap, seed=seed, img_name=img_name)
    elif phantoms == "aliens":
        generated = create_aliens(img_size=img_size, gray_values=gray_values,
                                    n=n, overlap=overlap, seed=seed, img_name=img_name)
    elif phantoms == "clouds":
        generated = create_clouds(img_size=img_size, gray_values=gray_values,
                                    n=n, overlap=overlap, seed=seed, img_name=img_name)
    elif phantoms == "paws":
        generated = create_paws(img_size=img_size, gray_values=gray_values,
                                    n=n, overlap=overlap, seed=seed, img_name=img_name)
    else:
        exit("please choose a valid class.")
    return generated

def create_image(x,y):
    """ Basic function required to create phantoms
    """
    image = np.ones([x,y])
    # Resize Image
    diag = len(np.diag(image)//2)
    if x == 80 and y == 80:
        plus = 8
    else: 
        plus = 10
    image = np.pad(image, pad_width=diag+plus)
    
    _ = np.linspace(-1, 1, image.shape[0])
    xv, yv = np.meshgrid(_,_)
    return image, xv, yv

def create_semilunars(img_size=512, gray_values=[80,120,180], n=1, 
                        overlap=False, seed=None, img_name=None):
    """ Create lunar like phantoms.

        Input:
            - img_size: defines the size of the image,
                should be 256 or 512.
            - gray_values: list with three values representing the 
                gray values to use for the images.
            - n: defines number of images to generate.
            - overlap: defines wether to make circles overlap or not
            - seed: defines the seed in order to have reproducable phantoms
            - img_name: string defining the path and filename to use.
                The filename should not have the extension, 
                it will be created as a png by default.
                Example of definition of img_name:
                    img_name = "dir/to/save/filename"

        Output:
            returns a list of phantoms. (Phantoms as numpy arrays)
    """
    # control correct image size
    if img_size == 512:
        x, y = 164, 164
    elif img_size == 256:
        x, y = 80, 80
    else:
        exit("img_size can only be set to 512 or 256")
    # control seed
    if seed != None:
        np.random.seed(seed)
    # define save paths
    if img_name != None:
        if not exists('data/'):
            makedirs('data/')
    # create phantoms
    semilunars = []
    for i in range(n):
        image, xv, yv = create_image(x,y)
        # define noise for randomization
        mu, sigma = 0 , 0.1
        noise =  np.abs(np.random.normal(mu, sigma, 2))
        # define shapes in image
        image[xv**2 + yv**2 >0.6] = 0
        image[xv**2 + yv**2 <0.49] = gray_values[1]
        if overlap == True:
            image[(xv-0.1+noise[0])**2 + (yv-0.1+noise[0])**2 <0.3] = gray_values[0]
            image[(xv-0.1+noise[1])**2 + (yv-0.1+noise[1])**2 <0.19] = gray_values[2]
            image[(xv-0.1+noise[0])**2 + (yv-0.1+noise[0])**2 <0.11] = gray_values[1]
            image[(xv-0.2+noise[1])**2 + (yv-0.2+noise[1])**2 <0.02] = gray_values[2]
        else: 
            image[(xv-0.1+noise[0])**2 + (yv-0.1+noise[0])**2 <0.3] = gray_values[0]
            image[(xv-0.1+noise[0])**2 + (yv-0.1+noise[0])**2 <0.19] = gray_values[2]
            image[(xv-0.1+noise[0])**2 + (yv-0.1+noise[0])**2 <0.11] = gray_values[1]
            image[(xv-0.2+noise[0])**2 + (yv-0.2+noise[0])**2 <0.02] = gray_values[2]
        semilunars.append(image.astype(np.uint8))
        # save image
        if img_name != None:
            Image.fromarray(image.astype(np.uint8)).save(f"{img_name}_{i}.png")
    return semilunars




def create_aliens(img_size=512, gray_values=[40,80,100], n=1, 
                overlap=False, seed=None, img_name=None):
    """ Create alien like phantoms.

        Input:
            - img_size: defines the size of the image,
                should be 256 or 512.
            - gray_values: list with three values representing the 
                gray values to use for the images.
            - n: defines number of images to generate.
            - overlap: defines wether to make circles overlap or not
            - seed: defines the seed in order to have reproducable phantoms
            - img_name: string defining the path and filename to use.
                The filename should not have the extension, 
                it will be created as a png by default.
                Example of definition of img_name:
                    img_name = "dir/to/save/filename"

        Output:
            returns a list of phantoms. (Phantoms as numpy arrays)
    """
    # control image size
    if img_size == 512:
        x, y = 164, 164
    elif img_size == 256:
        x, y = 80, 80
    else:
        exit("img_size can only be set to 512 or 256")
    # seed control
    if seed != None:
        np.random.seed(seed)
    # define save paths
    if img_name != None:
        if not exists('data/'):
            makedirs('data/')
    # create alien phantoms
    aliens = []
    for i in range(n):
        image, xv, yv = create_image(x,y)
        # create noise for randomization
        mu, sigma = 0 , 0.05
        noise =  np.abs(np.random.normal(mu, sigma, 2))
        # create phantom
        image[(xv-0.01)**2/0.5+(yv+0.01)**2>0.5] = 0
        image[(xv-0.01)**2/0.5+(yv+0.01)**2<0.48] = gray_values[2]
        if overlap == True:
            image[(xv+0.3-noise[0])**2/0.2+(yv+0.01-noise[0])**2<0.05] = gray_values[1]
            image[(xv+0.3-noise[1])**2/0.2+(yv+0.1-noise[1])**2<0.01] = gray_values[0]
            image[(xv-0.3+noise[0])**2/0.2+(yv+0.01+noise[0])**2<0.05] = gray_values[1]
            image[(xv-0.3+noise[1])**2/0.2+(yv+0.1+noise[1])**2<0.01] = gray_values[0]
            image[(xv-0.01-noise[0])**2+(yv-0.6+noise[0])**2/0.025<0.02] = gray_values[0]
        else: 
            image[(xv+0.3-noise[0])**2/0.2+(yv+0.01-noise[0])**2<0.05] = gray_values[1]
            image[(xv+0.3-noise[0])**2/0.2+(yv+0.1-noise[0])**2<0.01] = gray_values[0]
            image[(xv-0.3+noise[0])**2/0.2+(yv+0.01-noise[0])**2<0.05] = gray_values[1]
            image[(xv-0.3+noise[0])**2/0.2+(yv+0.1-noise[0])**2<0.01] = gray_values[0]
            image[(xv-0.01-noise[0])**2+(yv-0.6+noise[0])**2/0.025<0.02] = gray_values[0] 
        aliens.append(image.astype(np.uint8))
        # save image
        if img_name != None:
            Image.fromarray(image.astype(np.uint8)).save(f"{img_name}_{i}.png") 
    return aliens

def create_paws(img_size=512, gray_values=[80,120,180], n=1, 
                        overlap=False, seed=None, img_name=None):
    """ Create paw like phantoms.

        Input:
            - img_size: defines the size of the image,
                should be 256 or 512.
            - gray_values: list with three values representing the 
                gray values to use for the images.
            - n: defines number of images to generate.
            - overlap: defines wether to make circles overlap or not
            - seed: defines the seed in order to have reproducable phantoms
            - img_name: string defining the path and filename to use.
                The filename should not have the extension, 
                it will be created as a png by default.
                Example of definition of img_name:
                    img_name = "dir/to/save/filename"

        Output:
            returns a list of phantoms. (Phantoms as numpy arrays)
    """
    # img size control
    if img_size == 512:
        x, y = 164, 164
    elif img_size == 256:
        x, y = 80, 80
    else:
        exit("img_size can only be set to 512 or 256")
    # control seed
    if seed != None:
        np.random.seed(seed)
    # define save paths
    if img_name != None:
        if not exists('data/'):
            makedirs('data/')
    paws = []
    for i in range(n):
        image, xv, yv = create_image(x,y)

        thresh = np.random.choice(range(20),3)
        shift_x = np.random.choice(range(20,100))
        shift_y = np.random.choice(range(3))

        mu, sigma = 0 , 0.5
        noise =  np.abs(np.random.normal(mu, sigma, 2))

        # create images
        image[xv**2 + yv**2 >0.1] = 0
        image[xv**2 + yv**2 <0.09] = 255
        if img_size==256:
            image[25:231,30:40] = 255
            image[25:231, 216:226] = 255
            image[25:35, 40:216] = 255
            image[221:231, 40:216] = 255
            image[180:210+shift_y,50:80+shift_x] = 255
        else: 
            image[50:462,60:80] = 255
            image[50:462, 432:452] = 255
            image[50:70, 80:432] = 255
            image[442:462, 80:432] = 255
            image[360:420+shift_y,100:160+shift_x] = 255

        image[(xv-0.5)**2/0.2+(yv+0.3-noise[0])**2<0.05] = 255

        if thresh[0] <7: 
            image[(xv-0.006)**2/0.4+(yv+0.18)**2/0.7<0.01] = 0
        if thresh[1] > 5 : 
            image[(xv+0.17)**2/0.4+(yv+0.1)**2/0.7<0.01] = 0
        
        image[(xv-0.17)**2/0.4+(yv+0.1)**2/0.7<0.01] = 0
        image[(xv-0.01)**2/0.7+(yv-0.1)**2/0.7<0.02] = 0

        #fix the fucking pixels == 1
        image[image == 1] = 0
        
        paws.append(image.astype(np.uint8))
        # save image
        if img_name != None:
            Image.fromarray(image.astype(np.uint8)).save(f"{img_name}_{i}.png") 
    return paws


def create_clouds(img_size=512, gray_values=[80,120,180], n=1, 
                        overlap=False, seed=None, img_name=None):
    """ Create cloud like phantoms.

        Input:
            - img_size: defines the size of the image,
                should be 256 or 512.
            - gray_values: list with three values representing the 
                gray values to use for the images.
            - n: defines number of images to generate.
            - overlap: defines wether to make circles overlap or not
            - seed: defines the seed in order to have reproducable phantoms
            - img_name: string defining the path and filename to use.
                The filename should not have the extension, 
                it will be created as a png by default.
                Example of definition of img_name:
                    img_name = "dir/to/save/filename"

        Output:
            returns a list of phantoms. (Phantoms as numpy arrays)
    """
    # img size control
    if img_size == 512:
        x, y = 164, 164
    elif img_size == 256:
        x, y = 80, 80
    else:
        exit("img_size can only be set to 512 or 256")
    # control seed
    if seed != None:
        np.random.seed(seed)
    # define save paths
    if img_name != None:
        if not exists('data/'):
            makedirs('data/')


    clouds=[]
    for i in range(n):
        image, xv, yv = create_image(x,y)

        noise = np.random.normal(0,0.3,3)

        image[(xv-0.1)**2+(yv+0.01)**2/0.4<0.6] = 255
        image[(xv-0.1)**2+(yv+0.5)**2/0.4<0.1] = 255
        image[(xv-0.1)**2+(yv-0.5)**2/0.4<(0.1+noise[1])] = 255
        image[(xv+0.4)**2+(yv+0.5)**2/0.4<0.1] = 255
        image[(xv-0.4)**2+(yv-0.5)**2/0.4<(0.1+noise[1])] = 255
        image[(xv-0.6)**2+(yv+0.3)**2/0.4<(0.1-noise[2])] = 255
        image[(xv+0.4)**2+(yv-0.3)**2/0.4<(0.1-noise[2])] = 255
        
        n = np.random.normal(0,0.2,1)

        if 0.15<n[0]>0.3 :
            image[(xv-0.1)**2+(yv-0.1)**2/0.4<0.02] = 0
            image[(xv+0.4)**2+(yv-0.2)**2/0.4<0.005] = 0
            image[(xv-0.1)**2+(yv+0.2)**2/0.4<0.01] = 0
            image[(xv-0.4)**2+(yv+0.2)**2/0.4<0.02] = 0
            image[(xv+0.3)**2+(yv+0.3)**2/0.4<0.01] = 0
            image[(xv-0.4)**2+(yv-0.3)**2/0.4<0.02] = 0
        elif n[0]<0.09:
            image[(xv+0.4)**2+(yv-0.2)**2/0.4<0.005] = 0
            image[(xv-0.4)**2+(yv+0.2)**2/0.4<0.02] = 0
            image[(xv-0.4)**2+(yv-0.3)**2/0.4<0.02] = 0
            image[(xv+0.3)**2+(yv+0.3)**2/0.4<0.01] = 0
        else: 
            image[(xv+0.3)**2+(yv+0.3)**2/0.4<0.01] = 0
            image[(xv-0.4)**2+(yv-0.3)**2/0.4<0.02] = 0
            image[(xv-0.1)**2+(yv-0.1)**2/0.4<0.02] = 0

        clouds.append(image.astype(np.uint8))

        # save image
        if img_name != None:
            Image.fromarray(image.astype(np.uint8)).save(f"{img_name}_{i}.png") 
    return clouds